<?php

namespace Drupal\prometheus_metrics\Form;

use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Config\TypedConfigManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Routing\RouteBuilderInterface;
use Drupal\Core\Url;
use Drupal\prometheus_metrics\EventSubscriber\PrometheusDefaults;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Definition of prometheus_metrics route configration form.
 */
class MetricsConfigurationForm extends ConfigFormBase {
  /**
   * Used to rebuild route if route name provided.
   *
   * @var \Drupal\Core\Routing\RouteBuilderInterface
   */
  private $routeBuilder;

  /**
   * Constructs a \Drupal\system\ConfigFormBase object.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The factory for configuration objects.
   * @param \Drupal\Core\Routing\RouteBuilderInterface $routeBuilder
   *   Route builder service.
   * @param \Drupal\Core\Config\TypedConfigManagerInterface|null $typed_config_manager
   *   The typed config manager.
   */
  public function __construct(ConfigFactoryInterface $config_factory, RouteBuilderInterface $routeBuilder, ?TypedConfigManagerInterface $typed_config_manager = null) {
    if (version_compare(\Drupal::VERSION, '10.2.0', '>=')) {
      parent::__construct($config_factory, $typed_config_manager);
    }
    else {
      parent::__construct($config_factory);
    }
    $this->routeBuilder = $routeBuilder;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('router.builder'),
      $container->has('config.typed') ? $container->get('config.typed') : NULL
    );
  }

  /**
   * {@inheritDoc}
   */
  protected function getEditableConfigNames() {
    return [PrometheusDefaults::CONFIGURATION_NAME];
  }

  /**
   * {@inheritDoc}
   */
  public function getFormId() {
    return 'prometheus_metrics_route_configuration_form';
  }

  /**
   * {@inheritDoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $config = $this->config(PrometheusDefaults::CONFIGURATION_NAME);
    $form['metrics_route_path'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Metrics path'),
      '#description' => $this->t('Path to be used by Prometheus to collect metrics. Include preceding slash e.g. /manage/prometheus'),
      '#default_value' => $config->get('metrics_route_path') ?? ''
    ];

    $form['metrics_namespace'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Metrics namespace'),
      '#description' => $this->t('Namespace used for prometheus metrics. Should be unique to this service'),
      '#default_value' => $config->get('metrics_namespace') ?? ''
    ];

    $form['require_auth'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Require Authentication'),
      '#default_value' => $config->get('require_auth') ?? TRUE,
      '#description' => $this->t('If checked, authentication will be required to access the metrics endpoint. Add a user with the "access prometheus metrics" permission to the list of users who can access the metrics endpoint.'),
    ];

    // @todo make routes collected configurable, and review prom_php to see if possible to reduce memory usage.
    $form['metrics_route_memory'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Metrics Memory Ini Set'),
      '#description' => $this->t('(experimental) This module can collect a lot of metrics. Rendering metrics out can be expensive. Specify a memory limit for this route e.g. 256M. If not specified default ini values will be used.'),
      '#default_value' => $config->get('metrics_route_memory') ?? ''
    ];

    $form['metrics_exclude_routes'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Exclude routes'),
      '#description' => $this->t('Line separated list of routes to exclude. Wildcards may be used e.g. prometheus_*'),
      '#default_value' => is_array($config->get('metrics_exclude_routes')) ? join("\r\n",$config->get('metrics_exclude_routes')) : ''
    ];

    $form['metrics_consolidate_parameters'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Consolidate route parameters'),
      '#description' => $this->t('Line separated list of route parameters to consolidate. Format: route_name|parameter_name|replacement. Example: entity.node.canonical|node|{node}'),
      '#default_value' => is_array($config->get('metrics_consolidate_parameters')) ? join("\r\n",$config->get('metrics_consolidate_parameters')) : ''
    ];

    $form['reset_metrics'] = [
      '#type' => 'link',
      '#title' => $this->t('Reset all metrics. All metrics will be lost.'),
      '#url' => Url::fromRoute('prometheus_metrics.wipe'),
    ];
    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritDoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    $path = $form_state->getValue('metrics_route_path');

    if (!UrlHelper::isValid($path)) {
      $form_state->setErrorByName('metrics_route_path', $this->t('Path provided is invalid'));
    }
    // TODO: Change the autogenerated stub
    parent::validateForm($form, $form_state);
  }

  /**
   * {@inheritDoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $newRoute = $form_state->getValue('metrics_route_path');
    $oldRoute = $this->config(PrometheusDefaults::CONFIGURATION_NAME)
      ->get('metrics_route_path');

    $this->config(PrometheusDefaults::CONFIGURATION_NAME)
      ->set('metrics_namespace', $form_state->getValue('metrics_namespace'))
      ->set('require_auth', $form_state->getValue('require_auth'))
      ->save();

    $this->config(PrometheusDefaults::CONFIGURATION_NAME)
      ->set('metrics_route_memory', $form_state->getValue('metrics_route_memory'))
      ->save();

    $excludeRoutes = array_filter(
      array_unique(
        explode("\r\n", $form_state->getValue('metrics_exclude_routes'))
      ),
      function ($item){
        return !empty($item);
      }
    );
    $this->config(PrometheusDefaults::CONFIGURATION_NAME)
      ->set('metrics_exclude_routes', $excludeRoutes)
      ->save();

    $consolidateParameters = array_filter(
      array_unique(
        explode("\r\n", $form_state->getValue('metrics_consolidate_parameters'))
      ),
      function ($item){
        return !empty($item);
      }
    );
    $this->config(PrometheusDefaults::CONFIGURATION_NAME)
      ->set('metrics_consolidate_parameters', $consolidateParameters)
      ->save();

    if ($newRoute != $oldRoute) {
      $this->config(PrometheusDefaults::CONFIGURATION_NAME)
        ->set('metrics_route_path', $newRoute)
        ->save();
      $this->routeBuilder->rebuild();
    }
    parent::submitForm($form, $form_state);
  }

}
